import sys
import atexit
import sentry_sdk
import logging
from typing import Any
from functools import wraps
from importlib.resources import files
from sentry_sdk.integrations.logging import LoggingIntegration
from positron_common.user_config import user_config
from positron_common.common_dump import dump_data

# Load build env
build_env_resource = files('positron_common').joinpath('build_env')
build_env = "local"
if (build_env_resource.is_file()):
    try:
        build_env = build_env_resource.read_text().strip()
    except Exception as e:
        pass

# Initialize Sentry
sentry_sdk.init(
    dsn=user_config.sentry_dsn,
    # Set traces_sample_rate to 1.0 to capture 100%
    # of transactions for tracing.
    traces_sample_rate=1.0,
    # Set profiles_sample_rate to 1.0 to profile 100%
    # of sampled transactions.
    # We recommend adjusting this value in production.
    profiles_sample_rate=1.0,
    # Set environment so we can tell the difference between local and deployed
    environment=build_env,
    integrations=[LoggingIntegration(
        level=logging.DEBUG,         # Capture all info and above as breadcrumbs
        event_level=logging.ERROR,
    )],
  )

def track_command_usage(command_name) -> Any:
    """Decorator to track usage of commands in Sentry with arguments."""
    def decorator(func: Any) -> Any:
        @wraps(func)
        def wrapper(*args, **kwargs):
            with sentry_sdk.start_transaction(op="command", name=command_name):
                # Capture command usage event with arguments
                sentry_sdk.capture_event({
                    "message": f"Command Run: {command_name}",
                    "level": "info",
                    "tags": {"command": command_name},
                    "extra": {
                        "arguments": args,
                        "keyword_arguments": kwargs
                    }
                })
                # Execute the command function
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    sentry_sdk.capture_exception(e)
                    sentry_sdk.capture_event({
                        "message": f"Dump details from exception",
                        "level": "info",
                        "extra": dump_data(),
                    })
                    print(f"Trace ID: {sentry_sdk.get_current_span().trace_id}")
                    raise e
        return wrapper
    return decorator

def cleanup():
    # flush sentry events to ensure there's no sentry logging to the user
    sentry_sdk.flush()
    
atexit.register(cleanup)



import requests
from positron_common.user_config import user_config
from positron_common.env_config import env
from positron_common.cli.logging_config import logger

def is_auth_token_valid() -> bool:
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN}
    url = f'{env.API_BASE}/validate-user-auth-token'
    
    logger.debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)

    logger.debug(response)
    if response.status_code == 200:
        if response.json().get('sentryDsn'):
            user_config.sentry_dsn = response.json().get('sentryDsn')
            user_config.write()
        else:
            logger.warning('No sentryDsn in response')
        return True
    else:
        return False     

if __name__ == "__main__":
    print(is_auth_token_valid())



from pydantic import BaseModel, field_validator
from typing import Dict, Optional, List, Union
import os
import copy
import yaml
import re
from yaml import dump
from enum import Enum
from positron_common.utils import get_version, current_python_version
from positron_common.exceptions import RobbieException
from positron_common.user_config import user_config, REMOTE_FUNCTION_SECRET_KEY_NAME
from positron_common.constants import JOB_CONF_YAML_PATH, RUNTIME_ENV_PREFIX
from positron_common.cli.logging_config import logger
from positron_common.enums import JobRunType
from positron_common.cli.console import console
from positron_common.image_name import get_auto_image_name_and_cluster
from positron_common.job_api.funding_envs_images import *

class PositronJob(BaseModel):
    """
    The Job details as defined in the `python_job` from the `job_config.yaml` file.
    """
    name: Optional[str] = None
    funding_group_id: Optional[str] = None
    environment_id: Optional[str] = None
    image: Optional[str] = None # this is not written to the config file
    job_type: Optional[JobRunType] = None
    commands: Optional[List[str]] = None
    workspace_dir: Optional[str] = None
    max_tokens: Optional[int] = None
    max_time: Optional[int] = None
    env: Optional[Dict[str, str]] = None
    dependencies: Optional[str] = None
    conda_env: Optional[str] = None
    # below are values not writtent to the config file
    cluster: Optional[str] = None
    python_version: Optional[str] = current_python_version()
    robbie_sdk_version: Optional[str] = get_version()
    image_selection: Optional[str] = None
    funding_selection: Optional[str] = None
    environment_selection: Optional[str] = None

    @field_validator('commands', mode='after')
    def ensure_non_empty(cls, commands):
        return commands if len(commands) else None

    @field_validator('env', mode='after')
    def ensure_env_is_dict(cls, v):
        if isinstance(v, dict):
            return v
        raise ValueError('env must be a dictionary')

    @field_validator('max_time', mode='before')
    def ensure_max_time_is_int(cls, max_time: Union[int, str, None]) -> Union[int, None]:
        return cls._max_time_to_minutes(max_time)

    @field_validator('max_tokens', mode='before')
    def ensure_max_tokens_is_int(cls, max_tokens: Union[int, str, None]) -> Union[int, None]:
        return cls._max_tokens_to_int(max_tokens)

    def create_runtime_env(self) -> Dict[str, str]:
        """
        Used on the client side to create the prefixed runtime environment variables
        to avoid conflicts with the local environment variables.
        """
        env: Dict[str, str] = {}
        # without env_prefix?? yes, but maybe not in this file, but part of the deploy.py file.
        env[REMOTE_FUNCTION_SECRET_KEY_NAME] = user_config.user_auth_token
        if not self.env:
            return env
        for key, value in self.env.items():
            if (value == ""):
                env_var = os.environ.get(key)
                if env_var is None:
                    raise ValueError(f"The env prop {key} is unset inside job_config.yaml and also unset in local env vars. Please set this value.")
                env[f'{RUNTIME_ENV_PREFIX}{key}'] = env_var
            else:
                env[f'{RUNTIME_ENV_PREFIX}{key}'] = value
        return env

    @staticmethod
    def _max_time_to_minutes(max_time: Union[int, str, None]) -> Union[int, None]:
        if not max_time:
            return None
        if isinstance(max_time, int):
            return max_time
        matches = re.search(r'^(\d+):(\d{2})$', max_time)
        if matches is None:
            raise ValueError(f'Invalid Job Config: Field "max_time" ({max_time}) must have the format "HH:MM" or be a positive integer')
        try:
            hours = int(matches.group(1))
            minutes = int(matches.group(2))
        except:
            raise ValueError(f'Invalid Job Config: Field "max_time" ({max_time}) must have the format "HH:MM" or be a positive integer')
        if minutes >= 60:
            raise ValueError('Invalid Job Config: Field "max_time" ({max_time}) has invalid minutes! Must be 0 <= minutes < 60!')
        return hours * 60 + minutes

    @staticmethod
    def _max_tokens_to_int(max_tokens: Union[int, str, None]) -> Union[int, None]:
        if not max_tokens:
            return None
        if isinstance(max_tokens, int):
            return max_tokens
        try:
            max_tokens = int(max_tokens)
        except:
            raise ValueError(f'Invalid Job Config: "max_tokens" ({max_tokens}) needs to be a positive integer.')
        if max_tokens <= 0:
            raise ValueError(f'Invalid Job Config: "max_tokens" ({max_tokens}) needs to be a positive integer.')
        return max_tokens

    def validate_values(self) -> None:
        errors = []
        if self.env and not validate_env_vars(self.env):
            errors.append('At least one of the environment variables provided is invalid')
        if errors:
            raise RobbieException(f'Invalid configuration. Errors: {errors}')
        return None
    
    def namestr(self, obj, namespace):
        return [name for name in namespace if namespace[name] is obj]

    def to_string(self, title: str = 'None'):
        message = f"""
- python_version: {self.python_version}
- robbie_sdk_version: {self.robbie_sdk_version}
- name: {self.name}
- funding_group_id: {self.funding_group_id}
    - funding_selection: {self.funding_selection}
- environment_id: {self.environment_id}
    - environment_selection: {self.environment_selection}
- image: {self.image}
    - image_selection: {self.image_selection}
- cluster: {self.cluster}
- dependencies: {self.dependencies}
- conda_env: {self.conda_env}
- job_type: {self.job_type}
- workspace_dir: {self.workspace_dir}
- max_tokens: {self.max_tokens}
- max_time: {self.max_time}
- env: {self.env}
- commands: {self.commands}"""

        if title:
            return f"========== {title} ==========\n{message}"
        else:
            return message
        

class PositronJobConfig(BaseModel):
    """
    The `job_config.yaml` schema class.
    """
    version: float
    python_job: PositronJob

    def write_to_file(this, filename: str = JOB_CONF_YAML_PATH):
        copy_of_config = copy.deepcopy(this)
        del copy_of_config.python_job.robbie_sdk_version
        del copy_of_config.python_job.python_version
        del copy_of_config.python_job.cluster
        del copy_of_config.python_job.image_selection
        del copy_of_config.python_job.funding_selection
        del copy_of_config.python_job.environment_selection
        config_dict = copy_of_config.model_dump(
            exclude_unset=True
        )
        config_dict = convert_enums_to_values(config_dict)

        with open(filename, 'w') as file:
            file.write(dump(config_dict, sort_keys=False))


def convert_enums_to_values(d: dict) -> dict:
    """
    Converts Enum type values in the dictionary to their respective values.
    """
    for key, value in d.items():
        if isinstance(value, Enum):
            d[key] = value.value
        elif isinstance(value, dict):
            convert_enums_to_values(value)
    return d


def is_valid_key_value(keyvalue):
    """
    Validate that the key-value contains only alphanumeric characters, dashes, and underscores, and has no spaces.
    """
    return bool(re.match(r'^[\w-]+$', keyvalue))

def validate_env_vars(env_dict):
    """
    Validate the environment variables from the given dictionary.
    """
    valid = True
    for key, value in env_dict.items():
        if not is_valid_key_value(key):
            print(f"Invalid key (contains invalid characters or spaces): {key}")
            valid = False
        if value != "" and not is_valid_key_value(value):
            print(f"Invalid value (contains invalid characters or spaces): {value}")
            valid = False
    return valid

def merge_config(base_config: PositronJob, override_config: PositronJob) -> PositronJob:
    """
    Makes it easy to merge decorator configs on top of the YAML config.
    """
    update_data = override_config.dict(exclude_unset=True)
    updated_config = base_config.copy(update=update_data)
    return updated_config


def invalid_yaml_keys(yaml_job_keys) -> bool:
    """
    Validates the yaml keys against the PositronJob class.
    """
    # Get only attributes from PositronJob
    base_attrs = set(dir(BaseModel))
    derived_attrs = set(dir(PositronJob()))
    additional_attrs = derived_attrs - base_attrs

    # Exclude built-in attributes (e.g., __init__, __module__)
    validKeys = [attr for attr in additional_attrs if not attr.startswith('__')]
    
    weHaveInvalidYamlKeys = False
    for key in yaml_job_keys:
        if key not in validKeys:
            weHaveInvalidYamlKeys = True
            raise RobbieException(f'Error: wrong param in the job_config.yaml: -> {key}')

    return weHaveInvalidYamlKeys

def parse_job_config(config_path: str = 'job_config.yaml') -> Optional[PositronJob]:
    """
    Load the job configuration from the `job_config.yaml` file if it exists
    """
    
    if not os.path.exists(config_path):
        logger.debug('job_config.yaml file not found')
        return None
    
    try:
        # this happens sometimes and it causes the job to fail
        if os.path.getsize(config_path) == 0:
            logger.debug('job_config.yaml file is empty')
            return None
        with open(config_path, 'r') as job_config_file:
            job_config_dict = yaml.safe_load(job_config_file)
            job = job_config_dict["python_job"]
            if invalid_yaml_keys(job.keys()):
                return None
            job_config = PositronJobConfig(**job_config_dict)
            return job_config.python_job
    except Exception as e:
        print(f'Error loading job configuration! {str(e)}')
        return None

def merge_from_yaml_and_args(input_job_config: PositronJob, args_job_config: Union[PositronJob, None]) -> PositronJob:
    """
    Merge the job_config (from the yaml or empty object) file with the command line arguments for funding, environment, and image.
    Ensure that environment is in the funding group.

    Behavior:
    - Command line arguments take precedence over the job_config.yaml file
    - Command line arguments are not memorized in the job_config.yaml file 
    - Defaults are applied where selections are missing

    """

    funding_arg = args_job_config.funding_group_id if args_job_config else None
    environment_arg = args_job_config.environment_id if args_job_config else None
    image_arg = args_job_config.image if args_job_config else None
    
    logger.debug(input_job_config.to_string("Entering _merge_from_yaml_and_args()"))
    return_config = copy.deepcopy(input_job_config)
    logger.debug(f"return_config: {id(return_config)} created from input_job_config: {id(input_job_config)}")

    # Is there a funding group in the input config?
    if input_job_config.funding_group_id:
        logger.debug(f"return_config.funding_group_id: already set: {return_config.funding_group_id}")
        # ok the correct funding group is already in return_config.funding_group_id
        pass

    # Is there a funding group argument?
    if funding_arg:
        logger.debug(f"Setting or overriding return_config.funding_group_id: {return_config.funding_group_id} <-- funding_arg: {funding_arg}")
        return_config.funding_selection = "Overridden by argument"
        return_config.funding_group_id = funding_arg
        
    # pre-fetch the funding sources info and def env
    fs = list_funding_sources()
    if len(fs) == 0:
        logger.debug("No funding sources found.")
        return None

    def_env_id = None
    # Have we identified a funding group yet?
    if return_config.funding_group_id:
        logger.debug(f"Yes, we have a funding group: {return_config.funding_group_id}")
        if not return_config.environment_id:
            for _, val in fs.items():
                if (val[FS_ID] == return_config.funding_group_id):
                    def_env_id = val.get(FS_DEF_ENV_ID)
            if not def_env_id:
                console.print(f"[bold red]Can't get the default environment for the funding group: {return_config.funding_group_id}")
                return None   
            if not return_config.environment_id:
                logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- def_env_id: {def_env_id}")
                return_config.environment_id = def_env_id
                return_config.environment_selection = "Default"
        else:
            if not _env_in_funding_group(return_config.environment_id, return_config.funding_group_id):
                console.print(f"[bold red] Sorry, the environment in your job_config.yaml: {return_config.environment_id} is not the funding group: {return_config.funding_group_id}")
                return None
    else: 
        # No funding group id, let's get the user's personal and defaut environment
        logger.debug("Still no funding group id, let's get the user's personal one.")
        for _, val in fs.items():
            if (val[FS_TYPE] == FS_PERSONAL_TYPE):
                return_config.funding_group_id = val.get(FS_ID)
                def_env_id = val.get(FS_DEF_ENV_ID)
                logger.debug(f"Setting return_config.funding_group_id to PERSONAL: {return_config.funding_group_id}, fetched def_env_id: {def_env_id}")
        if return_config.funding_group_id:
            return_config.funding_selection = "Default"
        else:
            console.print("[bold red]Can't get the default funding source (personal).")
            return None
        if not def_env_id:
            console.print("[bold red]Can't get the default environment for the personal funding source.")
            return None
        # the user specified an environment in the job_config.yaml file
        if return_config.environment_id:
            # just for fun, lets check if its in the funding group
            if not _env_in_funding_group(return_config.environment_id, return_config.funding_group_id):
                console.print(f"[bold red] Sorry, the environment in your job_config.yaml: {return_config.environment_id} is not the funding group: {return_config.funding_group_id}")
                return None
        else:
            logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- def_env_id: {def_env_id}")
            return_config.environment_id = def_env_id
            return_config.environment_selection = "Default from Funding Source"
        
    # At this point we should have a funding group id and a default environment id
    if environment_arg:
        # check to make certain the environment argument is in the funding group
        if not _env_in_funding_group(environment_arg, return_config.funding_group_id):
            console.print(f"[bold red] Sorry, the environment you passed as an argument: {environment_arg} is not the funding group: {return_config.funding_group_id}")
            return None
        logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- environment_arg: {environment_arg}")
        return_config.environment_id = environment_arg
        return_config.environment_selection = "Overridden by argument"

    # Use configured image or auto-select
    if return_config.image and not image_arg:
            logger.debug(f'Using the image: {return_config.image} found in job_config.yaml. To override, please use the --image option.')
            return_config.image_selection = "From job_config.yaml"
    elif return_config.image and image_arg:
        logger.debug(f'overriding image job_config.yaml image: {return_config.image}, setting image to: {image_arg}')
        return_config.image = image_arg
        return_config.image_selection = "Overridden by argument"
    elif not return_config.image and image_arg:
        logger.debug(f'setting image job_config.yaml image: {return_config.image}, setting image to: {image_arg}')
        return_config.image = image_arg
        return_config.image_selection = "Set by argument"
    elif not return_config.image:
        logger.debug(f'No image in job_config.yaml file, setting to auto-select')
        return_config.image = "auto-select"
    
    # Handle special case
    if return_config.image == "auto-select":
        return_config.image, return_config.cluster = get_auto_image_name_and_cluster(return_config.funding_group_id, return_config.environment_id)
        return_config.image_selection = "Automatically Selected"
        
    logger.debug(return_config.to_string("Exiting _merge_from_yaml_and_args()"))

    return return_config

def _env_in_funding_group(env_id, funding_group_id):
    """ sanity check a environment id is in the funding group """
    envs = list_environments(funding_group_id)
    for _, val in envs.items():
        if (val[FS_ID] == env_id):
            return True
    return False

# if main then load and validate
if __name__ == "__main__":
    job_config = parse_job_config()
    if job_config:
        job_config.validate_values()
        print(job_config)



import sys
import os
import subprocess
from positron_common.cli.console import console, print_boxed_messages
from positron_common.build_env import build_env
from positron_common.utils import get_version
from positron_common.env_defaults import current
from positron_common.env_config import env
from positron_common.user_config import user_config
from positron_common.cli_args import args

def common_dump():
    console.print("Dumping Robbie local environment...")
    print_boxed_messages(sys.executable, "Python executable (sys.executable)")
    print_boxed_messages(os.__file__, "Python executable (os.__file__)")
    print_boxed_messages(get_version(), "Robbie Python SDK version - importlib.metadata.version('robbie')")
    result = subprocess.run(["pip", "show", "robbie"], capture_output=True, text=True)
    print_boxed_messages(result.stdout, "Robbie SDK version (% pip show robbie)")
    result = subprocess.run(["pip", "list"], capture_output=True, text=True)
    print_boxed_messages(result.stdout, "Installed Python Packages (% pip list)")
    if os.path.exists("./requirements.txt"):
        try:
            with open("./requirements.txt", 'r') as file:
                # Read the content of the file
                file_content = file.read()
                print_boxed_messages(file_content, "./requirements.txt")
        except Exception as e:
            print(f"An error occurred: {e}")
        
    print_boxed_messages(str(build_env), "build_env")
    print_boxed_messages(env.to_string(), "EnvConfig (env)")
    print_boxed_messages(current.to_string(), "EnvDefaults (current)")
    print_boxed_messages(user_config.to_string(), "User Config (user_config)")     
    print_boxed_messages(args.to_string(), "CLI Arguments (args)")

def dump_data():
    """All the same data as common_dump, but returned as a dict"""
    return {
        "Python executable (sys.executable)": sys.executable,
        "Python executable (os.__file__)": os.__file__,
        "Robbie Python SDK version - importlib.metadata.version('robbie')": get_version(),
        "Robbie SDK version (% pip show robbie)": subprocess.run(["pip", "show", "robbie"], capture_output=True, text=True).stdout,
        "Installed Python Packages (% pip list)": subprocess.run(["pip", "list"], capture_output=True, text=True).stdout,
        "./requirements.txt": open("./requirements.txt", 'r').read() if os.path.exists("./requirements.txt") else None,
        "build_env": str(build_env),
        "EnvConfig (env)": env.to_string(),
        "EnvDefaults (current)": current.to_string(),
        "User Config (user_config)": user_config.to_string(),
        "CLI Arguments (args)": args.to_string(),
    }


import asyncio
import socketio
import nest_asyncio
import json
import time
from rich.prompt import Confirm
from positron_common.env_config import env
from positron_common.exceptions import RobbieException
from positron_common.job_api.terminate_job import terminate_job
from ..cli.console import console, ROBBIE_BLUE
from ..cli.logging_config import logger
from positron_common.cli_args import args as cli_args

my_job_id = None

sio = socketio.AsyncClient(ssl_verify=False)
nest_asyncio.apply() # enabled nested event loops

@sio.event(namespace='/stdout-stream')
async def connect():
    console.print("Connected to your run's log stream!")

@sio.event(namespace='/stdout-stream')
async def message(message: str):
    try:
        log: dict = json.loads(message)
        level_name = log.get('log_level')
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S:%MS", time.gmtime(log.get('timestamp') / 1000))
        message = log.get('message')
        level_string = level_name.ljust(8)
        logger_name = log.get('app_name')

        if level_name == "INFO":
            level_string = f"[green]{level_name}[/green]"
        elif level_name == "DEBUG":
            level_string = f"[blue]{level_name}[/blue]"
        elif level_name == "ERROR":
            level_string = f"[red]{level_name}[/red]"
        elif level_name == "WARNING":
            level_string = f"[yellow]{level_name}[/yellow]"

        formatted_log = f"{logger_name}: {timestamp} {level_string}: {message}"
        console.print(formatted_log)
    except:
        console.print(log['message'], style=ROBBIE_BLUE)

@sio.event(namespace='/stdout-stream')
async def disconnect():
    logger.debug("Disconnecting from the log stream")
    pass

@sio.event(namespace='/stdout-stream')
async def error(err):
    console.print('An error occurred in the streaming process')
    logger.error(err)

async def start_stream(job_id: str):
    custom_headers = {
        "PositronAuthToken": env.USER_AUTH_TOKEN,
        "PositronJobId": job_id
    }
    await sio.connect(env.SOCKET_IO_DOMAIN, headers=custom_headers, socketio_path=env.SOCKET_IO_PATH)
    # TODO: I don't think we want to actually do this long term.
    try:
        await sio.wait()
    except asyncio.exceptions.CancelledError as error:
        # keyboard interrupt
        global my_job_id
        if my_job_id:
            if Confirm.ask("Interrupt received, do you want to terminate the run?", default=False):
                console.print("[yellow]Terminating run...[/yellow]")
                terminate_job(my_job_id, "User interrupted")
                # kind of a kluge, but we need to set this to false so it doesn't download results
                cli_args.download = False
            else:
                console.print("[yellow]Streaming ended. Please monitor run progress in the portal.[/yellow]")  
    finally:
        logger.debug("Done waiting...")

def start_stdout_stream(job_id: str):
    global my_job_id
    my_job_id = job_id 
    try:
        # Start the stream
        asyncio.get_event_loop().run_until_complete(start_stream(job_id))
    except Exception as error:
        raise RobbieException(error)





from typing import Optional
from positron_common.cli.console import console
from positron_common.utils import current_python_version

def validate_python_version(image: Optional[str]) -> bool:
    # example python version: 3.12
    # example image: robbie:1.0.0-gpu-py3.12-torch2.5-ubuntu22.04-dev
    if image:
        image_python_version = image.split("-")[2].replace("py", "")
        python_version = current_python_version()
        if image_python_version != python_version:
            console.print(f"[red]Error: Python version mismatch. The image you are using is built for Python {image_python_version}, but you are running Python {python_version}[/red]")
            return False
    return True



from positron_common.cli.console import console
from positron_common.cli.interactive import prompt_and_build_positron_job_config
from positron_common.constants import JOB_CONF_YAML_PATH
from positron_common.observability.main import track_command_usage

@track_command_usage("configure")
def configure() -> None:
    """
    Build a Robbie job configure file (job_config.yaml) interactively.

    """
    print("Please follow the prompts to build a 'job_config.yaml' in your current directory.")
    print("Values in brackets [] are default, use the <tab> key to see a menu of options.\n")
    
    cfg = prompt_and_build_positron_job_config()

    if cfg:
        cfg.write_to_file(filename=JOB_CONF_YAML_PATH)
        console.print(f"[green]Successfully wrote config file {JOB_CONF_YAML_PATH}'")
        console.print("File contents:")
        with open(JOB_CONF_YAML_PATH, 'r') as f:
            print(f.read())





    



import webbrowser
import typer
from typing_extensions import Annotated
from positron_common.user_config import user_config, DEFAULT_USER_CONFIG_PATH   
from positron_common.cli.logging_config import logger
from positron_common.cli.console import console
from positron_common.constants import APP_NAME, APP_HOME_DIR
from positron_common.auth_api.login import get_device_code_payload, get_user_auth_token, wait_for_access_token
from positron_common.job_api.validate_user_auth_token import is_auth_token_valid
from positron_common.env_config import env
from positron_common.exceptions import RemoteCallException
from positron_common.observability.main import track_command_usage


@track_command_usage("login")
def login(
    ctx: typer.Context = None,
    force: Annotated[bool, typer.Option("--force", help='Force the creation of a new API key')] = False,
) -> None:
    """
    Logs you in to your Robbie account and stores API key on your local machine.
    """
    if force or (not user_config.user_auth_token or not is_auth_token_valid()):
        # Get device code
        try:
            if not force:
                console.print('[red]Your Robbie authentication token is not valid, logging in.[/red]')

            console.print('Authenticating with Robbie...')
            device_code_data = get_device_code_payload()

            # Redirect to login
            console.print("1. If a browser window doesn't automatically launch, navigate to: ", device_code_data['verification_uri_complete'])
            console.print('2. Confirm the following code: ', device_code_data['user_code'])
            console.print('3. Enter your username and password!')
            console.print('')
            webbrowser.open(url=device_code_data['verification_uri_complete'], new=2, autoraise=True)

            # Wait for authentication
            access_token = wait_for_access_token(device_code_data)
            logger.debug(f'Access Token: {access_token}')

            # console.print('Requesting User Auth Token')
            user_token_response_data = get_user_auth_token(access_token)

            console.print(f'[green]Writing {APP_NAME} access token to: {DEFAULT_USER_CONFIG_PATH}')
            save_user_token(
                user_token_response_data['userAuthToken'],
                user_token_response_data['sentryDsn']
            )
            # ensure that "env" is updated with the new user token if it was imported previsouly
            env.USER_AUTH_TOKEN = user_config.user_auth_token
        except RemoteCallException as e:
            logger.debug(e, exc_info=True)
            console.print(f"[red]{e.user_friendly_message}")
            raise typer.Exit(code=1)
        except Exception as e:
            logger.debug(e, exc_info=True)
            console.print(f"[red]An error occurred: {e}. If the problem continues, reach out to our support team for help.\nEmail: support@robbie.run[/red]")
            raise typer.Exit(code=1)
    else:
        # was the command invoked from the CLI 
        if ctx:
            console.print(f'[green]You are good to go![/green]')
        

def save_user_token(user_token, sentry_dsn):
    user_config.user_auth_token = user_token
    user_config.sentry_dsn = sentry_dsn
    user_config.write()
    



import io
import os
import subprocess
import signal
import sys
import tempfile
import time
from importlib.resources import files
import ipywidgets as widgets
from ipywidgets import HBox, Layout
from nbformat import read, NO_CONVERT
from IPython import get_ipython
from IPython.display import display
from rich.jupyter import print
from positron_common.cli.logging_config import logger
from positron_common.cli.console import console, ROBBIE_DEFAULT, ROBBIE_BLUE
from positron_cli.run import run
from positron_common.job_api.funding_envs_images import *
from positron_common.cli.logging_config import set_log_level
from positron_cli.login import login
from positron_common.job_api.validate_user_auth_token import is_auth_token_valid
from positron_common.user_config import user_config
from positron_common.utils import _exit_by_mode, _nb

RUN_NAME_HELPER_TEXT = '(Optional customized run name)'
COMING_SOON_MENU_ITEM = 'Customize...(coming soon)'


def run_notebook(loglevel: str = None):
    """ 
    Run a notebook as a job on Robbie 
    
    This function will display a UI that allows the user to select the hardware to run the notebook on.
    
    """
    if loglevel:
        set_log_level(loglevel)

    if not _nb: # we are not in a notebook
        console.print("[red]Error: You are not running in a notebook.[/red]")
        logger.error("Error: User called run_notebook() while not running in a notebook.")
        return

    # check if the user is logged in
    login()

    #
    # Build the GUI
    #

    #
    # Left Column
    #
    left_column_title = widgets.HTML(
        value="<b>Robbie can run your notebook as a job!</b>",
    )

    # logo
    dog_img_resource = files('robbie').joinpath('img/dog.png')
    image = dog_img_resource.read_bytes()

    logo = widgets.Image(
        value=image,
        format='png',
        width=100,
        height=100,
    )

    style = {'description_width': 'initial'}

    #
    # Middle Column
    #
    middle_column_title = widgets.HTML(
        value="<b>Just choose your hardware!</b>",
    )
     # Environment Drop down
    _drop_dict = _load_envs_for_dropdown()
    if _drop_dict is None:
        console.print("[red] Error loading environments")
        return
    _drop_dict.update({COMING_SOON_MENU_ITEM: None})
    
    hardware_dropdown = widgets.Dropdown(
        options = _drop_dict.items(),
        # value=_drop_dict[1],
        description='Hardware Options:',
        style=style,
        layout={'width': 'max-content'}
    )

    _or_ = widgets.HTML(
        value="           - or - ",
    )

    # robbie does it for you
    """
    auto_checkbox = widgets.Checkbox(
        value=False,
        description='Let Robbie choose the best resources for you',
        disabled=False,
        indent=False
    )
    """

    # goes back to simple mode
    take_me_back_button = widgets.Button(
        description='Take me back',
        disabled=False,
        # button_style='info', # 'success', 'info', 'warning', 'danger' or ''
        tooltip='Go back to simple options',
        style = style
    )

    #
    # Right Column
    #
    right_column_title = widgets.HTML(
        value="<b>And get the ball rolling!</b>",
    )

    # optional name
    optional_name = widgets.Text(
        value=RUN_NAME_HELPER_TEXT,
        # description='Enter a name:',
        disabled=False,
        style=style
    )
    
    # "Run" button
    run_button = widgets.Button(
        description='Run',
        disabled=False,
        button_style='success', # 'success', 'info', 'warning', 'danger' or ''
        tooltip='Your notebook will be run as a job on Robbie',
        icon='play',
        style = style
    )

    # make sure the notebooke is saved
    saved_text = widgets.HTML(
        value="\n\nPlease ensure your notebook is saved before hitting to <b>Run</b> button",
    )

    # save help
    mac_save_img_resource = files('robbie').joinpath('img/mac_save.png')
    image = mac_save_img_resource.read_bytes()

    mac_short_cut_image = widgets.Image(
        value=image,
        format='png',
        width=30,
        height=21,
    )

    mac_shortcut_text = widgets.HTML(
        value=" on a Mac",
    )
    mac_save = HBox([mac_short_cut_image, mac_shortcut_text])
    # dont' stretch the image
    mac_save.layout.align_items = 'center'

    windows_save_img_resource = files('robbie').joinpath('img/windows_save.png')
    image = windows_save_img_resource.read_bytes()

    windows_shortcut_image = widgets.Image(
        value=image,
        format='png',
        width=49,
        height=18,
    )
    windows_shortcut_text = widgets.HTML(
        value=" on Windows",
    )

    coming_soon_text = widgets.HTML(
        value="Coming soon!",
    )

    # robbie does it for you
    tail_checkbox = widgets.Checkbox(
        value=False,
        description='Tail the remote machine stdout.',
        disabled=False,
        indent=False,
        style=style
    )
    windows_save = HBox([windows_shortcut_image, windows_shortcut_text])
    # dont' stretch the image
    windows_save.layout.align_items = 'center'

    save_hints = widgets.HBox([mac_save, windows_save])

    # layout
    left_column = widgets.VBox([left_column_title, logo])
    middle_column = widgets.VBox([middle_column_title, hardware_dropdown, tail_checkbox],layout=Layout(margin = '0 20px 0 20px'))
    right_column = widgets.VBox([right_column_title, optional_name, run_button, saved_text, save_hints], layout=Layout(margin = '0 20px 0 20px'))
    ui = widgets.HBox([left_column, middle_column, right_column])

    out = widgets.Output()
    # display everything
    display(ui, out)

 
    def on_take_me_back_button_clicked(change):
        """ When we display the full list of option (all disabled) this button gets gets us back to the simple mode"""
        with out:
            # print("on_take_me_back_clicked() - change:", change)
            middle_column.children = [middle_column_title, hardware_dropdown, tail_checkbox]
    
    take_me_back_button.on_click(on_take_me_back_button_clicked)
    

    # this is the hardware dropdown to change to customization mode
    def on_hw_dropdown_changed(change):
        """ This is experimental now to show users what is coming"""
        with out:
            # print("on_checkbox_clicked() - change:", change)
            if change.get('type') == 'change' and change.get('name') == 'value' and change.get('new') == COMING_SOON_MENU_ITEM:
                gpu_dropdown = widgets.Dropdown(
                    options=['Nvidia A10 (24GB)', 'Nvidia A100 (40GB)', 'Nvidia H100 (40GB)', 'CPU only'],
                    value='Nvidia A100 (40GB)',
                    description='GPU model:',
                    style=style,
                    layout={'width': 'max-content'},
                    disabled=True
                )
                cpu_slider=widgets.IntSlider(description='vCPU cores', min=1, max=64, step=1, value=4, style=style, disabled=True);
                memory_slider=widgets.IntSlider(description='System Memory (GB)', min=16, max=128, step=16, value=32, style=style, disabled=True);
                disksize_slider=widgets.IntSlider(description='Disk Size (GB)', min=16, max=128, step=16, value=32, style=style, disabled=True);
                middle_column.children = [middle_column_title, coming_soon_text, gpu_dropdown, cpu_slider, memory_slider, disksize_slider, take_me_back_button]

    hardware_dropdown.observe(on_hw_dropdown_changed)

    """
        # this is the robbie does it for you checkbox
        def on_checkbox_clicked(change):
            with out:
                # print("on_checkbox_clicked() - change:", change)
                if change.get('type') == 'change' and change.get('name') == 'value' and change.get('new') == True:
                    # print("Checkbox is checked")
                    # gpu_dropdown.disabled = True 
                    # cpu_slider.disabled = True
                    # memory_slider.disabled = True
                    hardware_dropdown.disabled = True
                    # disksize_slider.disabled = True
                elif change.get('type') == 'change' and change.get('name') == 'value' and change.get('new') == False:
                    # print("Checkbox is unchecked")
                    # gpu_dropdown.disabled = False
                    # cpu_slider.disabled = False
                    # memory_slider.disabled = False
                    hardware_dropdown.disabled = False
                    # disksize_slider.disabled = False

        # register the event handler
        auto_checkbox.observe(on_checkbox_clicked)
    """
    def on_button_clicked(b):
        """ The user clicked the run button """
        with out:
            logger.debug(f"Hardware dropdown value: {hardware_dropdown.value}")  
            if hardware_dropdown.value == None:
                console.print("[red] Please choose a hardware option, before running.")
                return
            
            py_path = _convert_notebook_to_py_based_on_platform()
            if py_path:
                console.print("[green]✔[/green] Converting notebook to Python..Success!")
            else:
                console.print("[red]Comverting notebook to Python..failed!")
                return
            
            console.print('Sending run to Robbie...')

            # this was the old way of doing things
            # kept here for reference
            '''
            run_cmd = f'robbie run "pip install uv && uv pip install -r ./requirements.txt && python {py_path}" --environment_id {hardware_dropdown.value} --y --auto-dep'
            if (optional_name.value and optional_name.value != RUN_NAME_HELPER_TEXT):
                run_cmd += f' --name "{optional_name.value}"'
            if tail_checkbox.value:
                run_cmd += ' --tail'
  
            # check if they enter a name
            rc = bash(run_cmd, print_stdout=True, print_stderr=True)
            if(rc != 0):
                console.print("[red] Job submission to Robbie..failed! Return code:", rc)
            '''
            _name = optional_name.value if optional_name.value != RUN_NAME_HELPER_TEXT else None

            run(name_arg=_name,
                funding_arg=None,
                environment_arg=hardware_dropdown.value,
                image_arg=None,
                commands_arg=f"pip install uv && uv pip install -r ./requirements.txt && python {py_path}",
                auto_capture_deps=True,
                tail=tail_checkbox.value,
                status=False,
                skip_prompts=True,
                interactive=False,
                download=False,
                create_only=False)
        
    # register the event handler
    run_button.on_click(on_button_clicked)

class VerboseCalledProcessError(subprocess.CalledProcessError):
    def __str__(self):
        if self.returncode and self.returncode < 0:
            try:
                msg = "Command '%s' died with %r." % (
                    self.cmd, signal.Signals(-self.returncode))
            except ValueError:
                msg = "Command '%s' died with unknown signal %d." % (
                    self.cmd, -self.returncode)
        else:
            msg = "Command '%s' returned non-zero exit status %d." % (
                self.cmd, self.returncode)

        return f'{msg}\n' \
               f'Stdout:\n' \
               f'{self.output}\n' \
               f'Stderr:\n' \
               f'{self.stderr}'


def bash(cmd, print_stdout=True, print_stderr=True):
    stdoutfile = tempfile.mktemp()
    stderrfile = tempfile.mktemp()
    with io.open(stdoutfile, 'w') as stdout, io.open(stderrfile, 'w') as stderr, io.open(stdoutfile, 'r') as all_stdout, io.open(stderrfile, 'r') as all_stderr:
        proc = subprocess.Popen(cmd, stderr=stderr, stdout=stdout, shell=True, universal_newlines=True,
                        executable='/bin/bash')

        stdout_text = ''
        stderr_text = ''
        while proc.poll() is None:
            stdout_line = all_stdout.read();
            stderr_line = all_stderr.read();
            stderr_text += stderr_line
            stdout_text += stdout_line
            sys.stdout.write(stdout_line)
            sys.stderr.write(stderr_line)
            time.sleep(0.5)
        
        stdout_line = all_stdout.read();
        stderr_line = all_stderr.read();
        stderr_text += stderr_line
        stdout_text += stdout_line
        sys.stdout.write(stdout_line)
        sys.stderr.write(stderr_line)

        if proc.wait() != 0:
            raise VerboseCalledProcessError(proc.returncode, cmd, stdout_text, stderr_text)
        return proc.returncode

def _convert_generic_notebook_to_py(nb_full_path):
    """
    Convert a generic notebook to a python file
    - works for VScode and Jupyter notebooks
    - remove things that are not compatible with generic python
    """
    if nb_full_path:
        console.print(f"Reading notebook file: {nb_full_path}")
        with open(nb_full_path) as fp:
            notebook = read(fp, NO_CONVERT)
        cells = notebook['cells']
        code_cells = [c for c in cells if c['cell_type'] == 'code']

        write_file = f"{os.path.splitext(os.path.basename(nb_full_path))[0]}.py"
        console.print(f"Writing Python file: {write_file}")
        with open(write_file, 'w') as fp1:
            for no_cell, cell in enumerate(code_cells):
                # print(f"####### Cell {no_cell} #########")
                # each cell is a string
                my_str = cell['source']

                # split the string into lines so we can look at each one
                list_of_lines = my_str.split('\n')
                clean_list = []
                for line in list_of_lines:
                    # discard magics
                    if line.startswith('%'):
                        # print("found %")
                        continue
                    if line.startswith("%%"):
                        # print("found %%")
                        continue
                    # discard robbie imports
                    if line.startswith("import robbie"):
                        continue
                    if line.startswith("robbie.login"):
                        continue
                    if line.startswith("robbie.init"):
                        continue
                    if line.startswith("robbie.run_notebook"):
                        continue
                    if line.startswith("!robbie login"):
                        continue
                    # wrap shell commands in os.system
                    if line.startswith("!"):
                        # print("found !")
                        line = line.replace(line[0], "", 1)
                        line = f"import os\nos.system('{line}')"
                    clean_list.append(line)
                my_str = '\n'.join(clean_list) 
                fp1.write(my_str)
                fp1.write("\n")
            #print("")
            return write_file
    return None
    
'''
Special way to convert a Google Colab notebook to a python file
- Google Colab notebooks have a different structure
- We need to use the Google Colab API to fetch the cells
'''
def _convert_colab_notebook_to_py(nb_full_path):
    # fetch the cells from the notebook frontend
    from google.colab import _message
    nb = _message.blocking_request('get_ipynb')
    if nb:
        cells = nb["ipynb"]['cells']
        code_cells = [c for c in cells if c['cell_type'] == 'code']

        print("nb_full_path:", nb_full_path)
        write_file = f"{os.path.splitext(os.path.basename(nb_full_path))[0]}.py"
        with open(write_file, 'w') as fp1:
            # go through the cells
            for no_cell, cell in enumerate(code_cells):
                print(f"####### Cell {no_cell} #########")
                print(cell['source'])
                # each cell is a list of strings
                list_of_lines = cell['source']
                # clean up the lines
                clean_list = []
                for line in list_of_lines:
                    # discard magics
                    if line.startswith('%'):
                        print("found %")
                        continue
                    if line.startswith("%%"):
                        print("found %%")
                        continue
                    # discard robbie imports
                    if line.startswith("import robbie"):
                        continue
                    if line.startswith("robbie.login"):
                        continue
                    if line.startswith("robbie.init"):
                        continue
                    if line.startswith("robbie.run_notebook"):
                        continue
                    if line.startswith("!robbie login"):
                        continue
                    # wrap shell commands in os.system
                    if line.startswith("!"):
                        print("found !")
                        line = line.replace(line[0], "", 1)
                        line = f"import os\nos.system('{line}')"
                    clean_list.append(line)
                my_str = '\n'.join(clean_list) 
                fp1.write(my_str)
                fp1.write("\n")
            return write_file
    return None

# Each notebook platform is unique in terms of how to derive the notebook name and how to convert its cells to a python file.
# We support three notebook types:
# 1. Visual Studio Code
# 2. Generic Jupyter Notebook
# 3. Google Colab


def _convert_notebook_to_py_based_on_platform():
    ip = get_ipython()
    if '__vsc_ipynb_file__' in ip.user_ns:
        console.print("[green]✔[/green] Visual Studio Code detected!") 
        return _convert_generic_notebook_to_py(ip.user_ns['__vsc_ipynb_file__'])
    elif 'JPY_SESSION_NAME' in os.environ:
        console.print("[green]✔[/green] Generic Jupyter Notebook detected!")
        return _convert_generic_notebook_to_py(os.getenv('JPY_SESSION_NAME'))
    elif 'google.colab' in str(ip):
        console.print("[green]✔[/green] Google Colab detected!")
        path = "/content/notebook.ipynb"
        _convert_colab_notebook_to_py(path)
    else:
        console.print("[red] Unanble to determine current notebook file name")
        return None

#
# Reads the environments from the PERSONAL funding group
#

def _load_envs_for_dropdown():
    """
    Load the environments for the dropdown list in the cell output
    """
    fs = list_funding_sources()
    if len(fs) == 0:
        return None
    
    personal_fs_id = None
    for key, val in fs.items():
        if (val[FS_TYPE] == FS_PERSONAL_TYPE):
            personal_fs_id = val.get(FS_ID)
    
    if personal_fs_id is None:
        return None
    envs = list_environments(personal_fs_id)
    if len(envs) == 0:
        return None
    
    dropdown_list = {}
    for _, val in envs.items():
        dropdown_list.update({val.get(ENV_NAME) + " - " + val.get(ENV_DESCRIPTION) + f' ({val.get(ENV_TPH)} tokens/hr)' : val.get(ENV_ID)})
    return dropdown_list





    

